

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of PSYCLONE version 2.0.3.210310 (10 March 2021)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2021
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  PSYCLONE 2.0 is released under the GNU Lesser General Public License 
//_/_/  (LGPL) version 3.0, and is subject to the following
//_/_/  
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/    
//_/_/  A commercial license for PSYCLONE 2.0 is also available to 
//_/_/  any interested party from the copyright holder.
//_/_/  
//_/_/  You can redistribute this copy of PSYCLONE 2.0 and/or modify it under 
//_/_/  the terms of the GNU General Public License as published by the Free 
//_/_/  Software Foundation, either version 3 of the License, or (at your 
//_/_/  option) any later version.
//_/_/  
//_/_/  You should have received a copy of the GNU General Public License
//_/_/  and the GNU Lesser General Public License along with PSYCLONE 2.0.  
//_/_/  If not, see <http://www.gnu.org/licenses/>.
//_/_/  
//_/_/  PSYCLONE 2.0 is distributed in the hope that it will be useful,
//_/_/  but WITHOUT ANY WARRANTY; without even the implied warranty of
//_/_/  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//_/_/  GNU General Public License for more details.
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#if !defined(_TIMESERIESCATALOG_H_)
#define _TIMESERIESCATALOG_H_

#include "PsyAPI.h"
#include "Bitmap.h"
#include "MessagePlayer.h"

namespace	cmlabs {

extern "C" {
	DllExport int8 TimeSeriesCatalog(PsyAPI* api);
}

struct TimeSeriesCatalogData {
	uint64 id;
	uint64 timestamp;
	char* data;
	uint32 size;
	std::string toJSON(uint32 datatype);
};


class TimeSeriesCatalogDataset {
public:
	TimeSeriesCatalogDataset() { reset(); }
	~TimeSeriesCatalogDataset() { cleanup(); }
	bool reset() {
		name.clear();
		shoulders.clear();
		color = Color();
		datatype = 0;
		lastID = 0;
		lastStored = 0;
		lastTimestamp = GetTimeNow();
		maxKeep = 0;
		shading = 0;
		minMSApart = 0;
		dataSeries.clear();
		return true;
	}
	bool cleanup() {
		std::list<TimeSeriesCatalogData>::iterator i = dataSeries.begin(), e = dataSeries.end();
		while (i != e) {
			delete (*i).data;
			i++;
		}
		return reset();
	}
	bool addData(uint64 t, const char* value);
	bool addData(uint64 t, int64 value);
	bool addData(uint64 t, float64 value);
	bool addData(uint64 t, uint64 time);
	bool addRawData(uint64 t, const char* data, uint32 size);
	bool repeatLastValue(uint64 t);
	DataMessage* retrieveData(uint64 sinceID = 0, uint64 sinceTime = 0, uint32 maxCount = 0, uint32 maxAgeMS = 0, uint32 minAgeMS = 0);
	std::string name;
	std::string shoulders;
	Color color;
	uint64 lastID;
	uint64 lastStored;
	uint64 lastTimestamp;
	uint32 maxKeep;
	uint32 minMSApart;
	uint32 datatype;
	uint32 shading;
	std::list<TimeSeriesCatalogData> dataSeries;
	utils::Mutex mutex;
};


class TimeSeriesCatalogEntry {
public:
	TimeSeriesCatalogEntry() { reset(); }
	~TimeSeriesCatalogEntry() { cleanup(); }
	bool reset() {
		dataset.clear();
		trigger.clear();
		key.clear();
		datatype = 0;
		lastTimestamp = GetTimeNow();
		return true;
	}
	bool cleanup() {
		return reset();
	}
	uint32 addData(DataMessage* msg, TimeSeriesCatalogDataset* dataset);
	std::string dataset;
	std::string trigger;
	std::string key;
	std::string timeKey;
	std::string durationKey;
	std::string value;
	uint64 lastTimestamp;
	uint32 datatype;
	utils::Mutex mutex;
};




} // namespace cmlabs

#endif // _TIMESERIESCATALOG_H_
